@@ -9,3 +9,9 @@ import _ "embed"
 
 //go:embed VERSION.txt
 var Version string
+
+// GoToolchainRev is the git hash from github.com/tailscale/go that this release
+// should be built using. It may end in a newline.
+//
+//go:embed go.toolchain.rev
+var GoToolchainRev string
